/*
	DooM MaP StaTistics, by Frans P. de Vries.

Derived from:

	DooM PostScript Maps Utility, by Frans P. de Vries.

And thus from:

	Doom Editor Utility, by Brendon Wyber and Raphaël Quinet.

	You are allowed to use any parts of this code in another program, as
	long as you give credits to the authors in the documentation and in
	the program itself.  Read the file README for more information.

	This program comes with absolutely no warranty.

	WADS.C - WAD files routines.
*/


/* the includes */
#include "dmmpst.h"

/* global variables */
WadPtr  WadFileList = NULL;  /* linked list of Wad files */
MDirPtr MasterDir   = NULL;  /* the master directory */


/*
	swap routines to handle big/little endian machines
*/
#ifdef DM_BIG_ENDIAN
/* big endian machines (eg. Sun, SGI) need to swap bytes within int's/long's */

void swapint( BCINT *i)
{
	BCINT t;

	((char *) &t)[0] = ((char *) i)[1];
	((char *) &t)[1] = ((char *) i)[0];
	*i = t;
}

void swaplong( BCLNG *l)
{
	BCLNG t;

	((char *) &t)[0] = ((char *) l)[3];
	((char *) &t)[1] = ((char *) l)[2];
	((char *) &t)[2] = ((char *) l)[1];
	((char *) &t)[3] = ((char *) l)[0];
	*l = t;
}

#else
/* little endian machines (eg. IBM PC) need to do nothing */
#endif


/*
	open the main WAD file, read in its directory and create the master directory
*/
void OpenMainWad( char *filename)
{
	MDirPtr lastp, newp;
	WadPtr wad;
	BCLNG n;

	/* open the WAD file */
	printf( "Loading main WAD file: %s...\n", filename);
	wad = BasicWadOpen( filename);
	if (strncmp( wad->type, "IWAD", 4))
		ProgError( "\"%s\" is not the main WAD file", filename);

	/* create the master directory */
	lastp = NULL;
	for (n = 0; n < wad->dirsize; n++)
	{
		newp = (MDirPtr) GetMemory( sizeof( struct MasterDirectory));
		newp->next = NULL;
		newp->wadfile = wad;
		memcpy( &(newp->dir), &(wad->directory[n]), sizeof( struct Directory));
		if (MasterDir)
			lastp->next = newp;
		else
			MasterDir = newp;
		lastp = newp;
	}

	/* check for expansion/ultimate/commercial/registered version */
	/* if you change this, bad things will happen to you... */
	if (FindMasterDir( MasterDir, "MAP58") != NULL) // Hexen DDC
		GameVersion = 0x08;
	else if (FindMasterDir( MasterDir, "E4M8") != NULL) // Doom ult, Heretic SSR
		GameVersion = 0x04;
	else if (FindMasterDir( MasterDir, "MAP28") != NULL) // Doom II, Strife, Hexen
		GameVersion = 0x02;
	else if (FindMasterDir( MasterDir, "E2M8") != NULL) // Doom reg, Heretic reg
		GameVersion = 0x01;
	/* else: shareware, demo */

	/* check for Doom end screen */
	if (FindMasterDir( MasterDir, "ENDOOM") != NULL)
		GameVersion += 0x00;
	/* check for Heretic end screen */
	else if (FindMasterDir( MasterDir, "ENDTEXT") != NULL)
		GameVersion += 0x10;
	/* check for Strife end screen */
	else if (FindMasterDir( MasterDir, "ENDSTRF") != NULL)
		GameVersion += 0x20;
	/* check for Hexen map lumps */
	else if (FindMasterDir( MasterDir, "BEHAVIOR") != NULL)
		GameVersion += 0x40;
	else
		ProgError( "\"%s\" is an unsupported main WAD file", filename);

	if (Verbose)
		printf( "Main WAD version: 0x%02x.\n", GameVersion);
}


/*
	open a patch WAD file, read in its directory and alter the master directory
*/
void OpenPatchWad( char *filename)
{
	MDirPtr mdir;
	WadPtr wad;
	BCINT n, l;
	char entryname[9];

	/* ignore the file if it doesn't exist */
	if (! Exists( filename))
	{
		printf( "Warning: patch WAD file \"%s\" doesn't exist.  Ignored.\n", filename);
		return;
	}

	/* open the WAD file */
	printf( "Loading patch WAD file: %s...\n", filename);
	wad = BasicWadOpen( filename);
	if (strncmp( wad->type, "PWAD", 4))
		ProgError( "\"%s\" is not a patch WAD file", filename);

	/* alter the master directory */
	l = 0;
	for (n = 0; n < wad->dirsize; n++)
	{
		strncpy( entryname, wad->directory[n].name, 8);
		entryname[8] = '\0';
		if (l == 0)
		{
			mdir = FindMasterDir( MasterDir, wad->directory[n].name);
			/* if this entry is not in the master directory, then add it */
			if (! mdir)
			{
				if (Verbose)
					printf( "   [Adding new entry %s]\n", entryname);
				mdir = MasterDir;
				while (mdir->next)
					mdir = mdir->next;
				mdir->next = (MDirPtr) GetMemory( sizeof( struct MasterDirectory));
				mdir = mdir->next;
				mdir->next = NULL;
			}
			/* if this is a level, then copy this entry and the next 10 (Hexen: 11) */
			else if (((GameVersion == 0x02 || GameVersion >= 0x20)
			          && wad->directory[n].name[0] == 'M'
			          && wad->directory[n].name[1] == 'A'
			          && wad->directory[n].name[2] == 'P'
			          && wad->directory[n].name[3] >= '0' && wad->directory[n].name[3] <= '9'
			          && wad->directory[n].name[4] >= '0' && wad->directory[n].name[4] <= '9'
			          && wad->directory[n].name[5] == '\0') ||
			         ((GameVersion != 0x02 && GameVersion < 0x20)
			          && wad->directory[n].name[0] == 'E'
			          && wad->directory[n].name[1] >= '0' && wad->directory[n].name[1] <= '9'
			          && wad->directory[n].name[2] == 'M'
			          && wad->directory[n].name[3] >= '0' && wad->directory[n].name[3] <= '9'
			          && wad->directory[n].name[4] == '\0'))
			{
				if (Verbose)
					printf( "   [Updating level %s]\n", entryname);
				l = (GameVersion & 0xF0) == 0x40 ? 11 : 10;
			}
			else
			{
				if (Verbose)
					printf( "   [Updating entry %s]\n", entryname);
			}
		}
		else
		{
			mdir = mdir->next;
			/* the level data should replace an existing level */
			if (mdir == NULL || strncmp( mdir->dir.name, wad->directory[n].name, 8))
				ProgError( "\\%s\" is not an understandable PWAD file (error with %s)", filename, entryname);
			l--;
		}
		mdir->wadfile = wad;
		memcpy( &(mdir->dir), &(wad->directory[n]), sizeof( struct Directory));
	}
}


/*
	close all the Wad files, deallocating the WAD file structures
*/
void CloseWadFiles()
{
	MDirPtr curd, nextd;
	WadPtr curw, nextw;

	/* close the Wad files */
	curw = WadFileList;
	WadFileList = NULL;
	while (curw)
	{
		nextw = curw->next;
		fclose( curw->fileinfo);
		FreeMemory( curw->directory);
		FreeMemory( curw);
		curw = nextw;
	}

	/* delete the master directory */
	curd = MasterDir;
	MasterDir = NULL;
	while (curd)
	{
		nextd = curd->next;
		FreeMemory( curd);
		curd = nextd;
	}
}


/*
	forget unused patch Wad files
*/
void CloseUnusedWadFiles()
{
	MDirPtr mdir;
	WadPtr curw, prevw;

	prevw = NULL;
	curw = WadFileList;
	while (curw)
	{
		/* check if the Wad file is used by a directory entry */
		mdir = MasterDir;
		while (mdir && mdir->wadfile != curw)
			mdir = mdir->next;
		if (mdir)
			prevw = curw;
		else
		{
			/* if this Wad file is never used, close it */
			if (prevw)
				prevw->next = curw->next;
			else
				WadFileList = curw->next;
			fclose( curw->fileinfo);
			FreeMemory( curw->directory);
			FreeMemory( curw);
		}
		curw = prevw->next;
	}
}


/*
	basic opening of WAD file and creation of node in Wad linked list
*/
WadPtr BasicWadOpen( char *filename)
{
	WadPtr curw, prevw;
	BCLNG i;

	/* find the WAD file in the Wad file list */
	prevw = WadFileList;
	if (prevw)
	{
		curw = prevw->next;
		while (curw && strcmp( filename, curw->filename))
		{
			prevw = curw;
			curw = prevw->next;
		}
	}
	else
		curw = NULL;

	/* if this entry doesn't exist, add it to the WadFileList */
	if (! curw)
	{
		curw = (WadPtr) GetMemory( sizeof( struct WadFileInfo));
		if (! prevw)
			WadFileList = curw;
		else
			prevw->next = curw;
		curw->next = NULL;
		curw->filename = filename;
	}

	/* open the file */
	if ((curw->fileinfo = fopen( filename, "rb")) == NULL)
	{
		if (! prevw)
			WadFileList = NULL;
		else
			prevw->next = curw->next;
		FreeMemory( curw);
		ProgError( "error opening \"%s\"", filename);
	}

	/* read in the WAD directory info */
	BasicWadRead( curw, curw->type, 4);
	if (strncmp( curw->type, "IWAD", 4) && strncmp( curw->type, "PWAD", 4))
		ProgError( "\"%s\" is not a valid WAD file", filename);
	BasicWadRead( curw, &curw->dirsize, sizeof( curw->dirsize));
	swaplong( &(curw->dirsize));
	BasicWadRead( curw, &curw->dirstart, sizeof( curw->dirstart));
	swaplong( &(curw->dirstart));

	/* read in the WAD directory itself */
	curw->directory = (DirPtr) GetMemory( curw->dirsize * sizeof( struct Directory));
	BasicWadSeek( curw, curw->dirstart);
	BasicWadRead( curw, curw->directory, sizeof( struct Directory) * curw->dirsize);
#ifdef DM_BIG_ENDIAN
	for (i = 0; i < curw->dirsize; i++)
	{
		DirPtr d = &(curw->directory[i]);
		swaplong( &(d->start));
		swaplong( &(d->size));
	}
#endif

	return curw;
}


/*
	read bytes from a file and store it into an address with error checking
*/
void BasicWadRead( WadPtr wadfile, void huge *addr, BCLNG size)
{
	if (fread( addr, 1, size, wadfile->fileinfo) != size)
		ProgError( "error reading from \"%s\"", wadfile->filename);
}


/*
	go to offset of a file with error checking
*/
void BasicWadSeek( WadPtr wadfile, BCLNG offset)
{
	if (fseek( wadfile->fileinfo, offset, 0))
		ProgError( "error reading from \"%s\"", wadfile->filename);
}


/*
	find an entry in the master directory
*/
MDirPtr FindMasterDir( MDirPtr from, char *name)
{
	while (from)
	{
		if (! strncmp( from->dir.name, name, 8))
			break;
		from = from->next;
	}
	return from;
}


/*
	check if a file exists and is readable
*/
Bool Exists( char *filename)
{
	FILE *test;

	if (! (test = fopen( filename, "rb")))
		return FALSE;
	fclose( test);
	return TRUE;
}

/* vim:set noexpandtab tabstop=2 softtabstop=2 shiftwidth=2: */
